@@ -1,4 +1,4 @@
-#     Copyright 2020, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -54,14 +54,6 @@ def main():
             if os.path.dirname(os.path.abspath(__file__)) != path_element
         ]
 
-    # For re-execution, we might not have done this.
-    from nuitka import Options  # isort:skip
-
-    Options.parseArgs()
-
-    # TODO: Stop using logging module, then this can be removed.
-    from nuitka import Tracing  # isort:skip
-
     # We don't care, and these are triggered by run time calculations of "range" and
     # others, while on python2.7 they are disabled by default.
 
@@ -71,12 +63,6 @@ def main():
     # not match, we restart ourselves with matching configuration.
     needs_reexec = False
 
-    # Inform the user about potential issues with the running version. e.g. unsupported.
-    from nuitka.PythonVersions import (
-        getSupportedPythonVersions,
-        python_version_str,
-    )
-
     if sys.flags.no_site == 0:
         needs_reexec = True
 
@@ -88,72 +74,25 @@ def main():
     if os.environ.get("PYTHONHASHSEED", "-1") != "0":
         needs_reexec = True
 
+    # For re-execution, we might not have done this.
+    from nuitka import Options  # isort:skip
+
     # In case we need to re-execute.
     if needs_reexec:
+        Options.parseArgs(will_reexec=True)
+
+        # TODO: If that's the only usage of Options, why do it at all, could start
+        # directly, and recognize the re-execution in the second run.
         if not Options.isAllowedToReexecute():
             sys.exit("Error, not allowed to re-execute, but that would be needed.")
 
-        our_filename = sys.modules[__name__].__file__
+        from nuitka.utils.ReExecute import reExecuteNuitka  # isort:skip
 
-        # Execute with full path as the process name, so it can find itself and its
-        # libraries.
-        args = [sys.executable, sys.executable]
+        reExecuteNuitka(pgo_filename=None)
 
-        if python_version_str >= "3.7" and sys.flags.utf8_mode:
-            args += ["-X", "utf8"]
+    Options.parseArgs(will_reexec=False)
 
-        args += ["-S", our_filename]
-
-        os.environ["NUITKA_BINARY_NAME"] = sys.modules["__main__"].__file__
-
-        if Options.is_nuitka_run:
-            args.append("--run")
-
-        # Same arguments as before.
-        args += sys.argv[1:] + list(Options.getMainArgs())
-
-        os.environ["NUITKA_PYTHONPATH"] = repr(sys.path)
-
-        from nuitka.importing.PreloadedPackages import (
-            detectPreLoadedPackagePaths,
-            detectPthImportedPackages,
-        )
-
-        os.environ["NUITKA_NAMESPACES"] = repr(detectPreLoadedPackagePaths())
-
-        if "site" in sys.modules:
-            os.environ["NUITKA_SITE_FILENAME"] = sys.modules["site"].__file__
-
-            os.environ["NUITKA_PTH_IMPORTED"] = repr(detectPthImportedPackages())
-
-        os.environ["NUITKA_SITE_FLAG"] = (
-            str(sys.flags.no_site) if "no_site" not in Options.getPythonFlags() else "1"
-        )
-
-        os.environ["PYTHONHASHSEED"] = "0"
-
-        from nuitka.utils import Execution  # isort:skip
-
-        Execution.callExec(args)
-
-    if python_version_str not in getSupportedPythonVersions():
-        # Do not disturb run of automatic tests, detected from the presence of
-        # that environment variable.
-        if "PYTHON" not in os.environ:
-            Tracing.general.warning(
-                "The version %r is not currently supported. Expect problems."
-                % python_version_str,
-            )
-
-    if os.name == "nt":
-        # Windows store Python's don't allow looking at the python, catch that.
-        try:
-            with open(sys.executable):
-                pass
-        except OSError:
-            sys.exit(
-                "Error, the Python from Windows store is not supported, check user manual."
-            )
+    Options.commentArgs()
 
     # Load plugins after we know, we don't execute again.
     from nuitka.plugins.Plugins import activatePlugins
@@ -193,4 +132,11 @@ def main():
 
 
 if __name__ == "__main__":
+    if "NUITKA_PACKAGE_HOME" in os.environ:
+        sys.path.insert(0, os.environ["NUITKA_PACKAGE_HOME"])
+
+        import nuitka  # just to have it loaded from there, pylint: disable=unused-import
+
+        del sys.path[0]
+
     main()
