@@ -1,4 +1,4 @@
-#     Copyright 2018, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -31,99 +31,73 @@ from .NodeBases import StatementChildrenHavingBase
 from .NodeMakingHelpers import (
     makeStatementExpressionOnlyReplacementNode,
     wrapExpressionWithNodeSideEffects,
-    wrapStatementWithSideEffects
+    wrapStatementWithSideEffects,
 )
-from .OperatorNodes import ExpressionOperationNOT
+from .OperatorNodesUnary import ExpressionOperationNot
+from .shapes.BuiltinTypeShapes import tshape_bool
 from .StatementNodes import StatementsSequence
 
 
 class ExpressionConditional(ExpressionChildrenHavingBase):
     kind = "EXPRESSION_CONDITIONAL"
 
-    named_children = (
-        "condition",
-        "expression_yes",
-        "expression_no"
-    )
+    named_children = ("condition", "expression_yes", "expression_no")
 
     def __init__(self, condition, expression_yes, expression_no, source_ref):
         ExpressionChildrenHavingBase.__init__(
             self,
-            values     = {
-                "condition"      : condition,
-                "expression_yes" : expression_yes,
-                "expression_no"  : expression_no
+            values={
+                "condition": condition,
+                "expression_yes": expression_yes,
+                "expression_no": expression_no,
             },
-            source_ref = source_ref
+            source_ref=source_ref,
         )
 
     def getBranches(self):
-        return (
-            self.getExpressionYes(),
-            self.getExpressionNo()
-        )
-
-    getExpressionYes = ExpressionChildrenHavingBase.childGetter(
-        "expression_yes"
-    )
-    getExpressionNo = ExpressionChildrenHavingBase.childGetter(
-        "expression_no"
-    )
-
-    getCondition = ExpressionChildrenHavingBase.childGetter(
-        "condition"
-    )
+        return (self.subnode_expression_yes, self.subnode_expression_no)
 
     def computeExpressionRaw(self, trace_collection):
         # This is rather complex stuff, pylint: disable=too-many-branches
 
         # Query the truth value after the expression is evaluated, once it is
         # evaluated in onExpression, it is known.
-        trace_collection.onExpression(
-            expression = self.getCondition()
-        )
-        condition = self.getCondition()
-
-        condition_may_raise = condition.mayRaiseException(BaseException)
-
-        if condition_may_raise:
-            trace_collection.onExceptionRaiseExit(
-                BaseException
-            )
+        condition = trace_collection.onExpression(expression=self.subnode_condition)
 
         # No need to look any further, if the condition raises, the branches do
         # not matter at all.
         if condition.willRaiseException(BaseException):
-            return condition, "new_raise", """\
+            return (
+                condition,
+                "new_raise",
+                """\
 Conditional expression already raises implicitly in condition, removing \
-branches."""
-
-        if not condition_may_raise and condition.mayRaiseExceptionBool(BaseException):
-            trace_collection.onExceptionRaiseExit(
-                BaseException
+branches.""",
             )
 
+        # Tell it we are evaluation it for boolean value only, it may demote
+        # itself possibly.
+        condition.computeExpressionBool(trace_collection)
+        condition = self.subnode_condition
+
         # Decide this based on truth value of condition.
         truth_value = condition.getTruthValue()
 
         # TODO: We now know that condition evaluates to true for the yes branch
         # and to not true for no branch, the branch should know that.
-        yes_branch = self.getExpressionYes()
+        yes_branch = self.subnode_expression_yes
 
         # Continue to execute for yes branch unless we know it's not going to be
         # relevant.
         if truth_value is not False:
             branch_yes_collection = TraceCollectionBranch(
-                parent = trace_collection,
-                name   = "conditional expression yes branch"
+                parent=trace_collection, name="conditional expression yes branch"
             )
 
-            branch_yes_collection.computeBranch(
-                branch = yes_branch
-            )
+            branch_yes_collection.computeBranch(branch=yes_branch)
 
             # May have just gone away, so fetch it again.
-            yes_branch = self.getExpressionYes()
+            yes_branch = self.subnode_expression_yes
 
             # If it's aborting, it doesn't contribute to merging.
             if yes_branch.willRaiseException(BaseException):
@@ -131,21 +105,18 @@ branches."""
         else:
             branch_yes_collection = None
 
-        no_branch = self.getExpressionNo()
+        no_branch = self.subnode_expression_no
 
         # Continue to execute for yes branch.
         if truth_value is not True:
             branch_no_collection = TraceCollectionBranch(
-                parent = trace_collection,
-                name   = "conditional expression no branch"
+                parent=trace_collection, name="conditional expression no branch"
             )
 
-            branch_no_collection.computeBranch(
-                branch = no_branch
-            )
+            branch_no_collection.computeBranch(branch=no_branch)
 
             # May have just gone away, so fetch it again.
-            no_branch = self.getExpressionNo()
+            no_branch = self.subnode_expression_no
 
             # If it's aborting, it doesn't contribute to merging.
             if no_branch.willRaiseException(BaseException):
@@ -159,46 +130,60 @@ branches."""
             trace_collection.replaceBranch(branch_no_collection)
         else:
             # Merge into parent execution.
-            trace_collection.mergeBranches(
-                branch_yes_collection,
-                branch_no_collection
-            )
+            trace_collection.mergeBranches(branch_yes_collection, branch_no_collection)
 
         if truth_value is True:
             return (
                 wrapExpressionWithNodeSideEffects(
-                    new_node = self.getExpressionYes(),
-                    old_node = condition
+                    new_node=self.subnode_expression_yes, old_node=condition
                 ),
                 "new_expression",
-                "Conditional expression predicted to yes case."
+                "Conditional expression predicted to yes case.",
             )
         elif truth_value is False:
             return (
                 wrapExpressionWithNodeSideEffects(
-                    new_node = self.getExpressionNo(),
-                    old_node = condition
+                    new_node=self.subnode_expression_no, old_node=condition
                 ),
                 "new_expression",
-                "Conditional expression predicted to no case."
+                "Conditional expression predicted to no case.",
             )
         else:
             return self, None, None
 
+    def computeExpressionDrop(self, statement, trace_collection):
+        result = makeStatementConditional(
+            condition=self.subnode_condition,
+            yes_branch=makeStatementExpressionOnlyReplacementNode(
+                expression=self.subnode_expression_yes, node=self.subnode_expression_yes
+            ),
+            no_branch=makeStatementExpressionOnlyReplacementNode(
+                expression=self.subnode_expression_no, node=self.subnode_expression_no
+            ),
+            source_ref=self.source_ref,
+        )
+
+        del self.parent
+
+        return (
+            result,
+            "new_statements",
+            """\
+Convert conditional expression with unused result into conditional statement.""",
+        )
+
     def mayHaveSideEffectsBool(self):
-        if self.getCondition().mayHaveSideEffectsBool():
+        # The bool will me made on either side.
+        if self.subnode_expression_yes.mayHaveSideEffectsBool():
             return True
 
-        if self.getExpressionYes().mayHaveSideEffectsBool():
-            return True
-
-        if self.getExpressionNo().mayHaveSideEffectsBool():
+        if self.subnode_expression_no.mayHaveSideEffectsBool():
             return True
 
         return False
 
     def mayRaiseException(self, exception_type):
-        condition = self.getCondition()
+        condition = self.subnode_condition
 
         if condition.mayRaiseException(exception_type):
             return True
@@ -206,91 +191,79 @@ branches."""
         if condition.mayRaiseExceptionBool(exception_type):
             return True
 
-        yes_branch = self.getExpressionYes()
+        yes_branch = self.subnode_expression_yes
 
         # Handle branches that became empty behind our back
-        if yes_branch is not None and \
-           yes_branch.mayRaiseException(exception_type):
+        if yes_branch is not None and yes_branch.mayRaiseException(exception_type):
             return True
 
-        no_branch = self.getExpressionNo()
+        no_branch = self.subnode_expression_no
 
         # Handle branches that became empty behind our back
-        if no_branch is not None and \
-           no_branch.mayRaiseException(exception_type):
+        if no_branch is not None and no_branch.mayRaiseException(exception_type):
             return True
 
         return False
 
     def mayRaiseExceptionBool(self, exception_type):
-        if self.getCondition().mayRaiseExceptionBool(exception_type):
+        if self.subnode_condition.mayRaiseExceptionBool(exception_type):
             return True
 
-        if self.getExpressionYes().mayRaiseExceptionBool(exception_type):
+        if self.subnode_expression_yes.mayRaiseExceptionBool(exception_type):
             return True
 
-        if self.getExpressionNo().mayRaiseExceptionBool(exception_type):
+        if self.subnode_expression_no.mayRaiseExceptionBool(exception_type):
             return True
 
         return False
 
+    def getIntegerValue(self):
+        result_yes = self.subnode_expression_yes.getIntegerValue()
+        result_no = self.subnode_expression_no.getIntegerValue()
+
+        if result_yes == result_no:
+            return result_yes
+        else:
+            return None
+
 
 class ExpressionConditionalBoolBase(ExpressionChildrenHavingBase):
-    named_children = (
-        "left",
-        "right"
-    )
+    named_children = ("left", "right")
 
     def __init__(self, left, right, source_ref):
         ExpressionChildrenHavingBase.__init__(
-            self,
-            values     = {
-                "left"  : left,
-                "right" : right,
-            },
-            source_ref = source_ref
+            self, values={"left": left, "right": right}, source_ref=source_ref
         )
 
-    getLeft = ExpressionChildrenHavingBase.childGetter(
-        "left"
-    )
-    getRight = ExpressionChildrenHavingBase.childGetter(
-        "right"
-    )
-
     def computeExpressionRaw(self, trace_collection):
         # Query the truth value after the expression is evaluated, once it is
         # evaluated in onExpression, it is known.
-        trace_collection.onExpression(
-            expression = self.getLeft()
-        )
-        left = self.getLeft()
+        left = trace_collection.onExpression(expression=self.subnode_left)
 
-        left_may_raise = left.mayRaiseException(BaseException)
-
-        if left_may_raise:
-            trace_collection.onExceptionRaiseExit(
-                BaseException
-            )
         # No need to look any further, if the condition raises, the branches do
         # not matter at all.
         if left.willRaiseException(BaseException):
-            return left, "new_raise", """\
+            return (
+                left,
+                "new_raise",
+                """\
 Conditional %s statements already raises implicitly in condition, removing \
-branches.""" % self.conditional_kind
-
-        if not left_may_raise and left.mayRaiseExceptionBool(BaseException):
-            trace_collection.onExceptionRaiseExit(
-                BaseException
+branches."""
+                % self.conditional_kind,
             )
 
+        if not left.mayRaiseException(BaseException) and left.mayRaiseExceptionBool(
+            BaseException
+        ):
+            trace_collection.onExceptionRaiseExit(BaseException)
+
         # Decide this based on truth value of condition.
         truth_value = left.getTruthValue()
 
         truth_value_use_left = self.conditional_kind == "or"
         truth_value_use_right = not truth_value_use_left
 
-        right = self.getRight()
+        right = self.subnode_right
 
         # Continue to execute for yes branch unless we know it's not going to be
         # relevant.
@@ -298,16 +271,14 @@ branches.""" % self.conditional_kind
             # TODO: We now know that left evaluates and we should tell the
             # branch that.
             branch_yes_collection = TraceCollectionBranch(
-                parent = trace_collection,
-                name   = "boolean %s right branch" % self.conditional_kind
+                parent=trace_collection,
+                name="boolean %s right branch" % self.conditional_kind,
             )
 
-            branch_yes_collection.computeBranch(
-                branch = right
-            )
+            branch_yes_collection.computeBranch(branch=right)
 
             # May have just gone away, so fetch it again.
-            right = self.getRight()
+            right = self.subnode_right
 
             # If it's aborting, it doesn't contribute to merging.
             if right.willRaiseException(BaseException):
@@ -315,34 +286,29 @@ branches.""" % self.conditional_kind
         else:
             branch_yes_collection = None
 
-
         if branch_yes_collection:
             # Merge into parent execution.
-            trace_collection.mergeBranches(
-                branch_yes_collection,
-                None
-            )
+            trace_collection.mergeBranches(branch_yes_collection, None)
 
         if truth_value is truth_value_use_left:
             return (
                 left,
                 "new_expression",
-                "Conditional '%s' expression predicted to left value." % self.conditional_kind
+                "Conditional '%s' expression predicted to left value."
+                % self.conditional_kind,
             )
         elif truth_value is truth_value_use_right:
             return (
-                wrapExpressionWithNodeSideEffects(
-                    new_node = right,
-                    old_node = left
-                ),
+                wrapExpressionWithNodeSideEffects(new_node=right, old_node=left),
                 "new_expression",
-                "Conditional '%s' expression predicted right value." % self.conditional_kind
+                "Conditional '%s' expression predicted right value."
+                % self.conditional_kind,
             )
         else:
             return self, None, None
 
     def mayRaiseException(self, exception_type):
-        left = self.getLeft()
+        left = self.subnode_left
 
         if left.mayRaiseException(exception_type):
             return True
@@ -350,7 +316,7 @@ branches.""" % self.conditional_kind
         if left.mayRaiseExceptionBool(exception_type):
             return True
 
-        right = self.getRight()
+        right = self.subnode_right
 
         if right.mayRaiseException(exception_type):
             return True
@@ -358,90 +324,112 @@ branches.""" % self.conditional_kind
         return False
 
     def mayRaiseExceptionBool(self, exception_type):
-        if self.getLeft().mayRaiseExceptionBool(exception_type):
+        # The and/or bool will be working on either side.
+        if self.subnode_left.mayRaiseExceptionBool(exception_type):
             return True
 
-        if self.getRight().mayRaiseExceptionBool(exception_type):
+        if self.subnode_right.mayRaiseExceptionBool(exception_type):
             return True
 
         return False
 
     def mayHaveSideEffectsBool(self):
-        if self.getLeft().mayHaveSideEffectsBool():
+        if self.subnode_left.mayHaveSideEffectsBool():
             return True
 
-        if self.getRight().mayHaveSideEffectsBool():
+        if self.subnode_right.mayHaveSideEffectsBool():
             return True
 
         return False
 
 
-class ExpressionConditionalOR(ExpressionConditionalBoolBase):
+class ExpressionConditionalOr(ExpressionConditionalBoolBase):
     kind = "EXPRESSION_CONDITIONAL_OR"
 
+    conditional_kind = "or"
+
     def __init__(self, left, right, source_ref):
         ExpressionConditionalBoolBase.__init__(
-            self,
-            left       = left,
-            right      = right,
-            source_ref = source_ref
+            self, left=left, right=right, source_ref=source_ref
         )
 
-        self.conditional_kind = "or"
+    def computeExpressionDrop(self, statement, trace_collection):
+        result = makeStatementConditional(
+            condition=self.subnode_left,
+            yes_branch=None,
+            no_branch=makeStatementExpressionOnlyReplacementNode(
+                expression=self.subnode_right, node=self.subnode_right
+            ),
+            source_ref=self.source_ref,
+        )
+
+        del self.parent
+
+        return (
+            result,
+            "new_statements",
+            """\
+Convert conditional 'or' expression with unused result into conditional statement.""",
+        )
 
 
-class ExpressionConditionalAND(ExpressionConditionalBoolBase):
+class ExpressionConditionalAnd(ExpressionConditionalBoolBase):
     kind = "EXPRESSION_CONDITIONAL_AND"
 
+    conditional_kind = "and"
+
     def __init__(self, left, right, source_ref):
         ExpressionConditionalBoolBase.__init__(
-            self,
-            left       = left,
-            right      = right,
-            source_ref = source_ref
+            self, left=left, right=right, source_ref=source_ref
         )
 
-        self.conditional_kind = "and"
+    def computeExpressionDrop(self, statement, trace_collection):
+        result = makeStatementConditional(
+            condition=self.subnode_left,
+            no_branch=None,
+            yes_branch=makeStatementExpressionOnlyReplacementNode(
+                expression=self.subnode_right, node=self.subnode_right
+            ),
+            source_ref=self.source_ref,
+        )
+
+        del self.parent
+
+        return (
+            result,
+            "new_statements",
+            """\
+Convert conditional 'and' expression with unused result into conditional statement.""",
+        )
 
 
 class StatementConditional(StatementChildrenHavingBase):
     kind = "STATEMENT_CONDITIONAL"
 
-    named_children = (
-        "condition",
-        "yes_branch",
-        "no_branch"
-    )
+    named_children = ("condition", "yes_branch", "no_branch")
 
     checkers = {
-        "yes_branch" : checkStatementsSequenceOrNone,
-        "no_branch" : checkStatementsSequenceOrNone,
+        "yes_branch": checkStatementsSequenceOrNone,
+        "no_branch": checkStatementsSequenceOrNone,
     }
 
     def __init__(self, condition, yes_branch, no_branch, source_ref):
         StatementChildrenHavingBase.__init__(
             self,
-            values     = {
-                "condition"  : condition,
-                "yes_branch" : yes_branch,
-                "no_branch"  : no_branch
+            values={
+                "condition": condition,
+                "yes_branch": yes_branch,
+                "no_branch": no_branch,
             },
-            source_ref = source_ref
+            source_ref=source_ref,
         )
 
-
-    getCondition = StatementChildrenHavingBase.childGetter("condition")
-    getBranchYes = StatementChildrenHavingBase.childGetter("yes_branch")
-    setBranchYes = StatementChildrenHavingBase.childSetter("yes_branch")
-    getBranchNo = StatementChildrenHavingBase.childGetter("no_branch")
-    setBranchNo = StatementChildrenHavingBase.childSetter("no_branch")
-
     def isStatementAborting(self):
-        yes_branch = self.getBranchYes()
+        yes_branch = self.subnode_yes_branch
 
         if yes_branch is not None:
             if yes_branch.isStatementAborting():
-                no_branch = self.getBranchNo()
+                no_branch = self.subnode_no_branch
 
                 if no_branch is not None:
                     return no_branch.isStatementAborting()
@@ -453,7 +441,7 @@ class StatementConditional(StatementChildrenHavingBase):
             return False
 
     def mayRaiseException(self, exception_type):
-        condition = self.getCondition()
+        condition = self.subnode_condition
 
         if condition.mayRaiseException(exception_type):
             return True
@@ -461,24 +449,22 @@ class StatementConditional(StatementChildrenHavingBase):
         if condition.mayRaiseExceptionBool(exception_type):
             return True
 
-        yes_branch = self.getBranchYes()
+        yes_branch = self.subnode_yes_branch
 
         # Handle branches that became empty behind our back
-        if yes_branch is not None and \
-           yes_branch.mayRaiseException(exception_type):
+        if yes_branch is not None and yes_branch.mayRaiseException(exception_type):
             return True
 
-        no_branch = self.getBranchNo()
+        no_branch = self.subnode_no_branch
 
         # Handle branches that became empty behind our back
-        if no_branch is not None and \
-           no_branch.mayRaiseException(exception_type):
+        if no_branch is not None and no_branch.mayRaiseException(exception_type):
             return True
 
         return False
 
     def needsFrame(self):
-        condition = self.getCondition()
+        condition = self.subnode_condition
 
         if condition.mayRaiseException(BaseException):
             return True
@@ -486,18 +472,16 @@ class StatementConditional(StatementChildrenHavingBase):
         if condition.mayRaiseExceptionBool(BaseException):
             return True
 
-        yes_branch = self.getBranchYes()
+        yes_branch = self.subnode_yes_branch
 
         # Handle branches that became empty behind our back
-        if yes_branch is not None and \
-           yes_branch.needsFrame():
+        if yes_branch is not None and yes_branch.needsFrame():
             return True
 
-        no_branch = self.getBranchNo()
+        no_branch = self.subnode_no_branch
 
         # Handle branches that became empty behind our back
-        if no_branch is not None and \
-           no_branch.needsFrame():
+        if no_branch is not None and no_branch.needsFrame():
             return True
 
         return False
@@ -505,96 +489,86 @@ class StatementConditional(StatementChildrenHavingBase):
     def computeStatement(self, trace_collection):
         # This is rather complex stuff, pylint: disable=too-many-branches,too-many-statements
 
-        trace_collection.onExpression(
-            expression = self.getCondition()
-        )
-        condition = self.getCondition()
-
-        condition_may_raise = condition.mayRaiseException(BaseException)
-
-        if condition_may_raise:
-            trace_collection.onExceptionRaiseExit(
-                BaseException
-            )
+        condition = trace_collection.onExpression(expression=self.subnode_condition)
 
         # No need to look any further, if the condition raises, the branches do
         # not matter at all.
         if condition.willRaiseException(BaseException):
             result = makeStatementExpressionOnlyReplacementNode(
-                expression = condition,
-                node       = self
+                expression=condition, node=self
             )
 
-            return result, "new_raise", """\
+            return (
+                result,
+                "new_raise",
+                """\
 Conditional statements already raises implicitly in condition, removing \
-branches."""
-
-        if not condition_may_raise and condition.mayRaiseExceptionBool(BaseException):
-            trace_collection.onExceptionRaiseExit(
-                BaseException
+branches.""",
             )
 
+        # Tell it we are evaluation it for boolean value only, it may demote
+        # itself possibly.
+        condition.computeExpressionBool(trace_collection)
+        condition = self.subnode_condition
+
         # Query the truth value after the expression is evaluated, once it is
         # evaluated in onExpression, it is known.
         truth_value = condition.getTruthValue()
 
         # TODO: We now know that condition evaluates to true for the yes branch
         # and to not true for no branch, the branch collection should know that.
-        yes_branch = self.getBranchYes()
-        no_branch = self.getBranchNo()
+        yes_branch = self.subnode_yes_branch
+        no_branch = self.subnode_no_branch
 
         # Handle branches that became empty behind our back.
         if yes_branch is not None:
-            if not yes_branch.getStatements():
+            if not yes_branch.subnode_statements:
                 yes_branch.finalize()
                 yes_branch = None
 
-                self.setBranchYes(None)
+                self.clearChild("yes_branch")
 
         if no_branch is not None:
-            if not no_branch.getStatements():
+            if not no_branch.subnode_statements:
                 no_branch.finalize()
                 no_branch = None
 
-                self.setBranchNo(None)
+                self.clearChild("no_branch")
 
         # Consider to not remove branches that we know won't be taken.
         if yes_branch is not None and truth_value is False:
             trace_collection.signalChange(
-                tags       = "new_statements",
-                source_ref = yes_branch.source_ref,
-                message    = "Removed conditional branch not taken due to false condition value."
+                tags="new_statements",
+                source_ref=yes_branch.source_ref,
+                message="Removed conditional branch that cannot be taken due to false condition value.",
             )
 
             yes_branch.finalize()
-            self.setBranchYes(None)
+            self.clearChild("yes_branch")
             yes_branch = None
 
         if no_branch is not None and truth_value is True:
             trace_collection.signalChange(
-                tags       = "new_statements",
-                source_ref = no_branch.source_ref,
-                message    = "Removed 'else' branch not taken due to true condition value."
+                tags="new_statements",
+                source_ref=no_branch.source_ref,
+                message="Removed 'else' branch that cannot be taken due to true condition value.",
             )
 
             no_branch.finalize()
-            self.setBranchNo(None)
+            self.clearChild("no_branch")
             no_branch = None
 
         # Continue to execute for yes branch unless we know it's not going to be
         # relevant.
         if yes_branch is not None:
             branch_yes_collection = TraceCollectionBranch(
-                parent = trace_collection,
-                name   = "conditional yes branch",
+                parent=trace_collection, name="conditional yes branch"
             )
 
-            branch_yes_collection.computeBranch(
-                branch = yes_branch
-            )
+            branch_yes_collection.computeBranch(branch=yes_branch)
 
             # May have just gone away, so fetch it again.
-            yes_branch = self.getBranchYes()
+            yes_branch = self.subnode_yes_branch
 
             # If it's aborting, it doesn't contribute to merging.
             if yes_branch is None or yes_branch.isStatementAborting():
@@ -605,16 +579,13 @@ branches."""
         # Continue to execute for yes branch.
         if no_branch is not None:
             branch_no_collection = TraceCollectionBranch(
-                parent = trace_collection,
-                name   = "conditional no branch"
+                parent=trace_collection, name="conditional no branch"
             )
 
-            branch_no_collection.computeBranch(
-                branch = no_branch
-            )
+            branch_no_collection.computeBranch(branch=no_branch)
 
             # May have just gone away, so fetch it again.
-            no_branch = self.getBranchNo()
+            no_branch = self.subnode_no_branch
 
             # If it's aborting, it doesn't contribute to merging.
             if no_branch is None or no_branch.isStatementAborting():
@@ -629,10 +600,7 @@ branches."""
             if branch_no_collection is not None:
                 trace_collection.replaceBranch(branch_no_collection)
         else:
-            trace_collection.mergeBranches(
-                branch_yes_collection,
-                branch_no_collection
-            )
+            trace_collection.mergeBranches(branch_yes_collection, branch_no_collection)
 
         # Both branches may have become empty, which case, the statement needs
         # not remain.
@@ -640,27 +608,33 @@ branches."""
             # Need to keep the boolean check.
             if truth_value is None:
                 condition = ExpressionBuiltinBool(
-                    value      = condition,
-                    source_ref = condition.getSourceReference()
+                    value=condition, source_ref=condition.getSourceReference()
                 )
 
             if condition.mayHaveSideEffects():
                 # With both branches eliminated, the condition remains as a side
                 # effect.
                 result = makeStatementExpressionOnlyReplacementNode(
-                    expression = condition,
-                    node       = self
+                    expression=condition, node=self
                 )
 
                 del self.parent
 
-                return result, "new_statements", """\
-Both branches have no effect, reduced to evaluate condition."""
+                return (
+                    result,
+                    "new_statements",
+                    """\
+Both branches have no effect, reduced to evaluate condition.""",
+                )
             else:
                 self.finalize()
 
-                return None, "new_statements", """\
-Removed conditional statement without effect."""
+                return (
+                    None,
+                    "new_statements",
+                    """\
+Removed conditional statement without effect.""",
+                )
 
         # Note: Checking the condition late, so that the surviving branch got
         # processed already. Returning without doing that, will corrupt the SSA
@@ -681,15 +655,20 @@ Removed conditional statement without effect."""
                     yes_branch.finalize()
 
             new_statement = wrapStatementWithSideEffects(
-                new_node   = new_statement,
-                old_node   = condition,
-                allow_none = True # surviving branch may empty
+                new_node=new_statement,
+                old_node=condition,
+                allow_none=True,  # surviving branch may empty
             )
 
             del self.parent
 
-            return new_statement, "new_statements", """\
-Condition for branch was predicted to be always %s.""" % choice
+            return (
+                new_statement,
+                "new_statements",
+                """\
+Condition for branch statement was predicted to be always %s."""
+                % choice,
+            )
 
         # If there is no "yes" branch, remove that. Maybe a bad idea though.
         if yes_branch is None:
@@ -698,29 +677,32 @@ Condition for branch was predicted to be always %s.""" % choice
             assert no_branch is not None
 
             new_statement = makeStatementConditional(
-                condition  = ExpressionOperationNOT(
-                    operand    = condition,
-                    source_ref = condition.getSourceReference()
+                condition=ExpressionOperationNot(
+                    operand=condition, source_ref=condition.getSourceReference()
                 ),
-                yes_branch = no_branch,
-                no_branch  = None,
-                source_ref = self.getSourceReference()
+                yes_branch=no_branch,
+                no_branch=None,
+                source_ref=self.source_ref,
             )
 
             del self.parent
 
-            return new_statement, "new_statements", """\
-Empty 'yes' branch for conditional statement treated with inverted condition check."""
+            return (
+                new_statement,
+                "new_statements",
+                """\
+Empty 'yes' branch for conditional statement treated with inverted condition check.""",
+            )
 
         return self, None, None
 
     def mayReturn(self):
-        yes_branch = self.getBranchYes()
+        yes_branch = self.subnode_yes_branch
 
         if yes_branch is not None and yes_branch.mayReturn():
             return True
 
-        no_branch = self.getBranchNo()
+        no_branch = self.subnode_no_branch
 
         if no_branch is not None and no_branch.mayReturn():
             return True
@@ -728,66 +710,70 @@ Empty 'yes' branch for conditional statement treated with inverted condition che
         return False
 
     def mayBreak(self):
-        yes_branch = self.getBranchYes()
+        yes_branch = self.subnode_yes_branch
 
         if yes_branch is not None and yes_branch.mayBreak():
             return True
 
-        no_branch = self.getBranchNo()
+        no_branch = self.subnode_no_branch
 
         if no_branch is not None and no_branch.mayBreak():
             return True
 
         return False
 
-
     def mayContinue(self):
-        yes_branch = self.getBranchYes()
+        yes_branch = self.subnode_yes_branch
 
         if yes_branch is not None and yes_branch.mayContinue():
             return True
 
-        no_branch = self.getBranchNo()
+        no_branch = self.subnode_no_branch
 
         if no_branch is not None and no_branch.mayContinue():
             return True
 
         return False
 
-    def getStatementNiceName(self):
+    @staticmethod
+    def getStatementNiceName():
         return "branch statement"
 
 
-def makeStatementConditional(condition, yes_branch, no_branch, source_ref):
-    """ Create conditional statement, with yes_branch not being empty.
+def makeNotExpression(expression):
+    # These are invertible with bool type shape.
+    if expression.isExpressionComparison() and expression.getTypeShape() is tshape_bool:
+        return expression.makeInverseComparision()
+    else:
+        return ExpressionOperationNot(
+            operand=expression, source_ref=expression.getSourceReference()
+        )
 
-        May have to invert condition to achieve that.
+
+def makeStatementConditional(condition, yes_branch, no_branch, source_ref):
+    """Create conditional statement, with yes_branch not being empty.
+
+    May have to invert condition to achieve that.
     """
 
     if yes_branch is None:
-        condition = ExpressionOperationNOT(
-            operand    = condition,
-            source_ref = condition.getSourceReference()
-        )
+        condition = makeNotExpression(expression=condition)
 
         yes_branch, no_branch = no_branch, yes_branch
 
     if yes_branch is not None and not yes_branch.isStatementsSequence():
         yes_branch = StatementsSequence(
-            statements = (yes_branch,),
-            source_ref = source_ref
+            statements=(yes_branch,), source_ref=yes_branch.source_ref
         )
 
     if no_branch is not None and not no_branch.isStatementsSequence():
         no_branch = StatementsSequence(
-            statements = (no_branch,),
-            source_ref = source_ref
+            statements=(no_branch,), source_ref=no_branch.source_ref
         )
 
     return StatementConditional(
-        condition  = condition,
-        yes_branch = yes_branch,
-        no_branch  = no_branch,
-        source_ref = source_ref
-
+        condition=condition,
+        yes_branch=yes_branch,
+        no_branch=no_branch,
+        source_ref=source_ref,
     )
