@@ -24,6 +24,8 @@ This is about collecting these constraints and to manage them.
 """
 
 import contextlib
+from collections import defaultdict
+from contextlib import contextmanager
 
 from nuitka import Tracing, Variables
 from nuitka.__past__ import iterItems  # Python3 compatibility.
@@ -35,14 +37,21 @@ from nuitka.nodes.shapes.BuiltinTypeShapes import tshape_dict
 from nuitka.nodes.shapes.StandardShapes import tshape_uninit
 from nuitka.tree.SourceReading import readSourceLine
 from nuitka.utils.FileOperations import relpath
-from nuitka.utils.InstanceCounters import counted_del, counted_init
+from nuitka.utils.InstanceCounters import (
+    counted_del,
+    counted_init,
+    isCountingInstances,
+)
 from nuitka.utils.ModuleNames import ModuleName
+from nuitka.utils.Timing import TimerReport
 
 from .ValueTraces import (
     ValueTraceAssign,
     ValueTraceDeleted,
     ValueTraceEscaped,
     ValueTraceInit,
+    ValueTraceInitStarArgs,
+    ValueTraceInitStarDict,
     ValueTraceLoopComplete,
     ValueTraceLoopIncomplete,
     ValueTraceMerge,
@@ -53,7 +62,61 @@ from .ValueTraces import (
 signalChange = None
 
 
-class CollectionStartpointMixin(object):
+@contextmanager
+def withChangeIndicationsTo(signal_change):
+    """Decide where change indications should go to."""
+
+    global signalChange  # Singleton, pylint: disable=global-statement
+
+    old = signalChange
+    signalChange = signal_change
+    yield
+    signalChange = old
+
+
+class CollectionUpdateMixin(object):
+    """Mixin to use in every collection to add traces."""
+
+    # Mixins are not allow to specify slots.
+    __slots__ = ()
+
+    def hasVariableTrace(self, variable, version):
+        return (variable, version) in self.variable_traces
+
+    def getVariableTrace(self, variable, version):
+        return self.variable_traces[(variable, version)]
+
+    def getVariableTraces(self, variable):
+        result = []
+
+        for key, variable_trace in iterItems(self.variable_traces):
+            candidate = key[0]
+
+            if variable is candidate:
+                result.append(variable_trace)
+
+        return result
+
+    def getVariableTracesAll(self):
+        return self.variable_traces
+
+    def addVariableTrace(self, variable, version, trace):
+        key = variable, version
+
+        assert key not in self.variable_traces, (key, self)
+        self.variable_traces[key] = trace
+
+    def addVariableMergeMultipleTrace(self, variable, traces):
+        version = variable.allocateTargetNumber()
+
+        trace_merge = ValueTraceMerge(traces)
+
+        self.addVariableTrace(variable, version, trace_merge)
+
+        return version
+
+
+class CollectionStartpointMixin(CollectionUpdateMixin):
     """Mixin to use in start points of collections.
 
     These are modules, functions, etc. typically entry points.
@@ -148,23 +211,6 @@ class CollectionStartpointMixin(object):
     def getExceptionRaiseCollections(self):
         return self.exception_collections
 
-    def hasVariableTrace(self, variable, version):
-        return (variable, version) in self.variable_traces
-
-    def getVariableTrace(self, variable, version):
-        return self.variable_traces[(variable, version)]
-
-    def getVariableTraces(self, variable):
-        result = []
-
-        for key, variable_trace in iterItems(self.variable_traces):
-            candidate = key[0]
-
-            if variable is candidate:
-                result.append(variable_trace)
-
-        return result
-
     def hasEmptyTraces(self, variable):
         # TODO: Combine these steps into one for performance gains.
         traces = self.getVariableTraces(variable)
@@ -175,42 +221,38 @@ class CollectionStartpointMixin(object):
         traces = self.getVariableTraces(variable)
         return areReadOnlyTraces(traces)
 
-    def getVariableTracesAll(self):
-        return self.variable_traces
-
-    def addVariableTrace(self, variable, version, trace):
-        key = variable, version
-
-        assert key not in self.variable_traces, (key, self)
-        self.variable_traces[key] = trace
-
-    def addVariableMergeMultipleTrace(self, variable, traces):
-        version = variable.allocateTargetNumber()
-
-        trace_merge = ValueTraceMerge(traces)
-
-        self.addVariableTrace(variable, version, trace_merge)
-
-        return version
-
     def initVariableUnknown(self, variable):
         trace = ValueTraceUnknown(owner=self.owner, previous=None)
 
-        self.addVariableTrace(variable=variable, version=0, trace=trace)
+        self.addVariableTrace(variable, 0, trace)
 
         return trace
 
     def _initVariableInit(self, variable):
         trace = ValueTraceInit(self.owner)
 
-        self.addVariableTrace(variable=variable, version=0, trace=trace)
+        self.addVariableTrace(variable, 0, trace)
+
+        return trace
+
+    def _initVariableInitStarArgs(self, variable):
+        trace = ValueTraceInitStarArgs(self.owner)
+
+        self.addVariableTrace(variable, 0, trace)
+
+        return trace
+
+    def _initVariableInitStarDict(self, variable):
+        trace = ValueTraceInitStarDict(self.owner)
+
+        self.addVariableTrace(variable, 0, trace)
 
         return trace
 
     def _initVariableUninit(self, variable):
         trace = ValueTraceUninit(owner=self.owner, previous=None)
 
-        self.addVariableTrace(variable=variable, version=0, trace=trace)
+        self.addVariableTrace(variable, 0, trace)
 
         return trace
 
@@ -247,6 +289,7 @@ class CollectionStartpointMixin(object):
 
     def initVariable(self, variable):
         if variable.isParameterVariable():
+            # TODO: That's not happening, maybe just assert against it.
             result = self._initVariableInit(variable)
         elif variable.isLocalVariable():
             result = self._initVariableUninit(variable)
@@ -294,7 +337,8 @@ class TraceCollectionBase(object):
 
     __slots__ = ("owner", "parent", "name", "value_states", "variable_actives")
 
-    __del__ = counted_del()
+    if isCountingInstances():
+        __del__ = counted_del()
 
     @counted_init
     def __init__(self, owner, name, parent):
@@ -346,13 +390,27 @@ class TraceCollectionBase(object):
     def markActiveVariableAsEscaped(self, variable):
         current = self.getVariableCurrentTrace(variable=variable)
 
+        if not current.isEscapeOrUnknownTrace():
+            version = variable.allocateTargetNumber()
+
+            self.addVariableTrace(
+                variable,
+                version,
+                ValueTraceEscaped(owner=self.owner, previous=current),
+            )
+
+            self.markCurrentVariableTrace(variable, version)
+
+    def markActiveVariableAsUnknown(self, variable):
+        current = self.getVariableCurrentTrace(variable=variable)
+
         if not current.isUnknownTrace():
             version = variable.allocateTargetNumber()
 
             self.addVariableTrace(
-                variable=variable,
-                version=version,
-                trace=ValueTraceEscaped(owner=self.owner, previous=current),
+                variable,
+                version,
+                ValueTraceUnknown(owner=self.owner, previous=current),
             )
 
             self.markCurrentVariableTrace(variable, version)
@@ -373,18 +431,25 @@ class TraceCollectionBase(object):
             result = ValueTraceLoopComplete(loop_node, current, shapes)
 
         version = variable.allocateTargetNumber()
-        self.addVariableTrace(variable=variable, version=version, trace=result)
+        self.addVariableTrace(variable, version, result)
 
         self.markCurrentVariableTrace(variable, version)
 
         return result
 
+    def markActiveVariablesAsEscaped(self):
+        for variable in self.getActiveVariables():
+            if variable.isTempVariable():
+                continue
+
+            self.markActiveVariableAsEscaped(variable)
+
     def markActiveVariablesAsUnknown(self):
         for variable in self.getActiveVariables():
             if variable.isTempVariable():
                 continue
 
-            self.markActiveVariableAsEscaped(variable)
+            self.markActiveVariableAsUnknown(variable)
 
     @staticmethod
     def signalChange(tags, source_ref, message):
@@ -392,7 +457,22 @@ class TraceCollectionBase(object):
         signalChange(tags, source_ref, message)
 
     def onUsedModule(self, module_name, module_relpath):
-        return self.parent.onUsedModule(module_name, module_relpath)
+        return self.parent.onUsedModule(
+            module_name=module_name, module_relpath=module_relpath
+        )
+
+    def onUsedFunction(self, function_body):
+        owning_module = function_body.getParentModule()
+
+        # Make sure the owning module is added to the used set. This is most
+        # important for helper functions, or modules, which otherwise have
+        # become unused.
+        addUsedModule(owning_module)
+
+        needs_visit = owning_module.addUsedFunction(function_body)
+
+        if needs_visit:
+            function_body.computeFunctionRaw(self)
 
     @staticmethod
     def mustAlias(a, b):
@@ -415,18 +495,26 @@ class TraceCollectionBase(object):
         # to be considered escaped, pylint: disable=unused-argument
 
         for variable in self.getActiveVariables():
+            # TODO: Move this to the variable, and prepare and cache it better for
+            # compile time savings.
             if variable.isModuleVariable():
-                # print variable
-
-                self.markActiveVariableAsEscaped(variable)
+                self.markActiveVariableAsUnknown(variable)
 
             elif variable.isLocalVariable():
-                if variable.hasAccessesOutsideOf(self.owner) is not False:
+                if (
+                    str is not bytes
+                    and variable.hasWritersOutsideOf(self.owner) is not False
+                ):
+                    self.markActiveVariableAsUnknown(variable)
+                elif variable.hasAccessesOutsideOf(self.owner) is not False:
                     self.markActiveVariableAsEscaped(variable)
 
     def removeKnowledge(self, node):
         if node.isExpressionVariableRef():
-            self.markActiveVariableAsEscaped(node.variable)
+            if node.variable.isModuleVariable():
+                self.markActiveVariableAsUnknown(node.variable)
+            else:
+                self.markActiveVariableAsEscaped(node.variable)
 
     def onValueEscapeStr(self, node):
         # TODO: We can ignore these for now.
@@ -435,18 +523,6 @@ class TraceCollectionBase(object):
     def removeAllKnowledge(self):
         self.markActiveVariablesAsUnknown()
 
-    def getVariableTrace(self, variable, version):
-        return self.parent.getVariableTrace(variable, version)
-
-    def hasVariableTrace(self, variable, version):
-        return self.parent.hasVariableTrace(variable, version)
-
-    def addVariableTrace(self, variable, version, trace):
-        self.parent.addVariableTrace(variable, version, trace)
-
-    def addVariableMergeMultipleTrace(self, variable, traces):
-        return self.parent.addVariableMergeMultipleTrace(variable, traces)
-
     def onVariableSet(self, variable, version, assign_node):
         variable_trace = ValueTraceAssign(
             owner=self.owner,
@@ -454,7 +530,7 @@ class TraceCollectionBase(object):
             previous=self.getVariableCurrentTrace(variable=variable),
         )
 
-        self.addVariableTrace(variable=variable, version=version, trace=variable_trace)
+        self.addVariableTrace(variable, version, variable_trace)
 
         # Make references point to it.
         self.markCurrentVariableTrace(variable, version)
@@ -466,12 +542,14 @@ class TraceCollectionBase(object):
         # remember the delete of the current
         old_trace = self.getVariableCurrentTrace(variable)
 
+        # TODO: Annotate value content as escaped.
+
         variable_trace = ValueTraceDeleted(
             owner=self.owner, del_node=del_node, previous=old_trace
         )
 
         # Assign to not initialized again.
-        self.addVariableTrace(variable=variable, version=version, trace=variable_trace)
+        self.addVariableTrace(variable, version, variable_trace)
 
         # Make references point to it.
         self.markCurrentVariableTrace(variable, version)
@@ -589,6 +667,30 @@ class TraceCollectionBase(object):
         else:
             return expression, change_tags, change_desc
 
+    def computedExpressionResultRaw(self, expression, change_tags, change_desc):
+        """Make sure the replacement expression is computed.
+
+        Use this when a replacement expression needs to be seen by the trace
+        collection and be computed, without causing any duplication, but where
+        otherwise there would be loss of annotated effects.
+
+        This may e.g. be true for nodes that need an initial run to know their
+        exception result and type shape.
+
+        This is for raw, i.e. subnodes are not yet computed automatically.
+        """
+
+        # Need to compute the replacement still.
+        new_expression = expression.computeExpressionRaw(self)
+
+        if new_expression[0] is not expression:
+            # Signal intermediate result as well.
+            self.signalChange(change_tags, expression.getSourceReference(), change_desc)
+
+            return new_expression
+        else:
+            return expression, change_tags, change_desc
+
     def mergeBranches(self, collection_yes, collection_no):
         """Merge two alternative branches into this trace.
 
@@ -664,34 +766,39 @@ class TraceCollectionBase(object):
             self.replaceBranch(collections[0])
             return None
 
-        variable_versions = {}
+        # print("Enter mergeMultipleBranches", len(collections))
+        with TimerReport(
+            message="Running merge for %s took %%.2f seconds" % collections,
+            decider=lambda: 0,
+        ):
+            variable_versions = defaultdict(OrderedSet)
 
-        for collection in collections:
-            for variable, version in iterItems(collection.variable_actives):
-                if variable not in variable_versions:
-                    variable_versions[variable] = OrderedSet((version,))
-                else:
+            for collection in collections:
+                for variable, version in iterItems(collection.variable_actives):
                     variable_versions[variable].add(version)
 
-        for collection in collections:
+            for collection in collections:
+                for variable, versions in iterItems(variable_versions):
+                    if variable not in collection.variable_actives:
+                        versions.add(0)
+
+            self.variable_actives = {}
+
             for variable, versions in iterItems(variable_versions):
-                if variable not in collection.variable_actives:
-                    versions.add(0)
+                if len(versions) == 1:
+                    (version,) = versions
+                else:
+                    version = self.addVariableMergeMultipleTrace(
+                        variable=variable,
+                        traces=tuple(
+                            self.getVariableTrace(variable, version)
+                            for version in versions
+                        ),
+                    )
 
-        self.variable_actives = {}
+                self.markCurrentVariableTrace(variable, version)
 
-        for variable, versions in iterItems(variable_versions):
-            if len(versions) == 1:
-                (version,) = versions
-            else:
-                version = self.addVariableMergeMultipleTrace(
-                    variable=variable,
-                    traces=tuple(
-                        self.getVariableTrace(variable, version) for version in versions
-                    ),
-                )
-
-            self.markCurrentVariableTrace(variable, version)
+            # print("Leave mergeMultipleBranches", len(collections))
 
     def replaceBranch(self, collection_replace):
         self.variable_actives.update(collection_replace.variable_actives)
@@ -746,12 +853,14 @@ class TraceCollectionBase(object):
     def onLocalsDictEscaped(self, locals_scope):
         self.parent.onLocalsDictEscaped(locals_scope)
 
-    def getCompileTimeComputationResult(self, node, computation, description):
+    def getCompileTimeComputationResult(
+        self, node, computation, description, user_provided=False
+    ):
         new_node, change_tags, message = getComputationResult(
             node=node,
             computation=computation,
             description=description,
-            user_provided=False,
+            user_provided=user_provided,
         )
 
         if change_tags == "new_raise":
@@ -778,8 +887,8 @@ class TraceCollectionBase(object):
         self.parent.addOutlineFunction(outline)
 
 
-class TraceCollectionBranch(TraceCollectionBase):
-    __slots__ = ()
+class TraceCollectionBranch(CollectionUpdateMixin, TraceCollectionBase):
+    __slots__ = ("variable_traces",)
 
     def __init__(self, name, parent):
         TraceCollectionBase.__init__(self, owner=parent.owner, name=name, parent=parent)
@@ -787,6 +896,9 @@ class TraceCollectionBranch(TraceCollectionBase):
         # Detach from others
         self.variable_actives = dict(parent.variable_actives)
 
+        # For quick access without going to parent.
+        self.variable_traces = parent.variable_traces
+
     def computeBranch(self, branch):
         if branch.isStatementsSequence():
             result = branch.computeStatementsSequence(trace_collection=self)
@@ -803,9 +915,6 @@ class TraceCollectionBranch(TraceCollectionBase):
 
         return variable_trace
 
-    def onLocalsDictEscaped(self, locals_scope):
-        return self.parent.onLocalsDictEscaped(locals_scope)
-
     def dumpTraces(self):
         Tracing.printSeparator()
         self.parent.dumpTraces()
@@ -849,10 +958,22 @@ class TraceCollectionFunction(CollectionStartpointMixin, TraceCollectionBase):
         )
 
         if function_body.isExpressionFunctionBody():
-            for parameter_variable in function_body.getParameters().getAllVariables():
+            parameters = function_body.getParameters()
+
+            for parameter_variable in parameters.getTopLevelVariables():
                 self._initVariableInit(parameter_variable)
                 self.variable_actives[parameter_variable] = 0
 
+            list_star_variable = parameters.getListStarArgVariable()
+            if list_star_variable is not None:
+                self._initVariableInitStarArgs(list_star_variable)
+                self.variable_actives[list_star_variable] = 0
+
+            dict_star_variable = parameters.getDictStarArgVariable()
+            if dict_star_variable is not None:
+                self._initVariableInitStarDict(dict_star_variable)
+                self.variable_actives[dict_star_variable] = 0
+
         for closure_variable in function_body.getClosureVariables():
             self.initVariableUnknown(closure_variable)
             self.variable_actives[closure_variable] = 0
@@ -868,6 +989,23 @@ class TraceCollectionFunction(CollectionStartpointMixin, TraceCollectionBase):
                 function_body.locals_scope = None
 
 
+class TraceCollectionPureFunction(TraceCollectionFunction):
+    """Pure functions don't feed their parent."""
+
+    __slots__ = ("used_functions",)
+
+    def __init__(self, function_body):
+        TraceCollectionFunction.__init__(self, parent=None, function_body=function_body)
+
+        self.used_functions = OrderedSet()
+
+    def getUsedFunctions(self):
+        return self.used_functions
+
+    def onUsedFunction(self, function_body):
+        self.used_functions.add(function_body)
+
+
 class TraceCollectionModule(CollectionStartpointMixin, TraceCollectionBase):
     __slots__ = (
         "variable_versions",
@@ -905,7 +1043,7 @@ class TraceCollectionModule(CollectionStartpointMixin, TraceCollectionBase):
 def areEmptyTraces(variable_traces):
     """Do these traces contain any writes or accesses."""
     # Many cases immediately return, that is how we do it here,
-    # pylint: disable=too-many-return-statements
+    # pylint: disable=too-many-branches,too-many-return-statements
 
     for variable_trace in variable_traces:
         if variable_trace.isAssignTrace():
@@ -927,6 +1065,11 @@ def areEmptyTraces(variable_traces):
                 # Checking definite is enough, the merges, we shall see
                 # them as well.
                 return False
+        elif variable_trace.isEscapeTrace():
+            if variable_trace.getUsageCount():
+                # Checking definite is enough, the merges, we shall see
+                # them as well.
+                return False
         elif variable_trace.isMergeTrace():
             if variable_trace.getUsageCount():
                 # Checking definite is enough, the merges, we shall see
@@ -958,6 +1101,8 @@ def areReadOnlyTraces(variable_traces):
             pass
         elif variable_trace.isUnknownTrace():
             return False
+        elif variable_trace.isEscapeTrace():
+            pass
         elif variable_trace.isMergeTrace():
             pass
         elif variable_trace.isLoopTrace():
