@@ -32,6 +32,7 @@ Values can be seen as:
 * LoopComplete (complete knowledge of loop types)
 """
 
+from nuitka.nodes.shapes.BuiltinTypeShapes import tshape_dict, tshape_tuple
 from nuitka.nodes.shapes.StandardShapes import (
     ShapeLoopCompleteAlternative,
     ShapeLoopInitialAlternative,
@@ -142,6 +143,14 @@ class ValueTraceBase(object):
     def isUnknownTrace():
         return False
 
+    @staticmethod
+    def isEscapeTrace():
+        return False
+
+    @staticmethod
+    def isEscapeOrUnknownTrace():
+        return False
+
     @staticmethod
     def isMergeTrace():
         return False
@@ -173,6 +182,22 @@ class ValueTraceBase(object):
     def getTruthValue():
         return None
 
+    @staticmethod
+    def getComparisonValue():
+        return None
+
+    @staticmethod
+    def getAttributeNode():
+        return None
+
+    @staticmethod
+    def getAttributeNodeTrusted():
+        return None
+
+    @staticmethod
+    def getAttributeNodeVeryTrusted():
+        return None
+
 
 class ValueTraceUnassignedBase(ValueTraceBase):
     __slots__ = ()
@@ -254,6 +279,26 @@ class ValueTraceInit(ValueTraceBase):
         return False
 
 
+class ValueTraceInitStarArgs(ValueTraceInit):
+    @staticmethod
+    def getTypeShape():
+        return tshape_tuple
+
+    @staticmethod
+    def hasShapeDictionaryExact():
+        return False
+
+
+class ValueTraceInitStarDict(ValueTraceInit):
+    @staticmethod
+    def getTypeShape():
+        return tshape_dict
+
+    @staticmethod
+    def hasShapeDictionaryExact():
+        return True
+
+
 class ValueTraceUnknown(ValueTraceBase):
     __slots__ = ()
 
@@ -285,6 +330,10 @@ class ValueTraceUnknown(ValueTraceBase):
     def isUnknownTrace():
         return True
 
+    @staticmethod
+    def isEscapeOrUnknownTrace():
+        return True
+
     @staticmethod
     def mustHaveValue():
         return False
@@ -293,6 +342,20 @@ class ValueTraceUnknown(ValueTraceBase):
     def mustNotHaveValue():
         return False
 
+    def getAttributeNode(self):
+        # TODO: Differentiate unknown with not previous node from ones with for performance and
+        # clarity.
+        if self.previous is not None:
+            return self.previous.getAttributeNodeVeryTrusted()
+
+    def getAttributeNodeTrusted(self):
+        if self.previous is not None:
+            return self.previous.getAttributeNodeVeryTrusted()
+
+    def getAttributeNodeVeryTrusted(self):
+        if self.previous is not None:
+            return self.previous.getAttributeNodeVeryTrusted()
+
 
 class ValueTraceEscaped(ValueTraceUnknown):
     __slots__ = ()
@@ -313,6 +376,36 @@ class ValueTraceEscaped(ValueTraceUnknown):
         if self.merge_usage_count <= 2:
             self.previous.addMergeUsage()
 
+    def mustHaveValue(self):
+        return self.previous.mustHaveValue()
+
+    def mustNotHaveValue(self):
+        return self.previous.mustNotHaveValue()
+
+    def getReplacementNode(self, usage):
+        return self.previous.getReplacementNode(usage)
+
+    @staticmethod
+    def isUnknownTrace():
+        return False
+
+    @staticmethod
+    def isEscapeTrace():
+        return True
+
+    @staticmethod
+    def isEscapeOrUnknownTrace():
+        return True
+
+    def getAttributeNode(self):
+        return self.previous.getAttributeNodeTrusted()
+
+    def getAttributeNodeTrusted(self):
+        return self.previous.getAttributeNodeTrusted()
+
+    def getAttributeNodeVeryTrusted(self):
+        return self.previous.getAttributeNodeVeryTrusted()
+
 
 class ValueTraceAssign(ValueTraceBase):
     __slots__ = ("assign_node", "replace_it")
@@ -365,6 +458,28 @@ class ValueTraceAssign(ValueTraceBase):
     def getTruthValue(self):
         return self.assign_node.subnode_source.getTruthValue()
 
+    def getComparisonValue(self):
+        return self.assign_node.subnode_source.getComparisonValue()
+
+    def getAttributeNode(self):
+        return self.assign_node.subnode_source
+
+    def getAttributeNodeTrusted(self):
+        source_node = self.assign_node.subnode_source
+
+        if source_node.hasShapeTrustedAttributes():
+            return source_node
+        else:
+            return None
+
+    def getAttributeNodeVeryTrusted(self):
+        source_node = self.assign_node.subnode_source
+
+        if source_node.isExpressionImportModuleHard():
+            return source_node
+        else:
+            return None
+
 
 class ValueTraceMergeBase(ValueTraceBase):
     """Merge of two or more traces or start of loops."""
@@ -477,6 +592,11 @@ class ValueTraceMerge(ValueTraceMergeBase):
         # Now all agreed and were not unknown, so we can conclude all false or all true.
         return any_true
 
+    def getComparisonValue(self):
+        # TODO: Support multiple values as candidates, e.g. both 1, 3 could be compared to 2, for
+        # now we are delaying that.
+        return None
+
 
 class ValueTraceLoopBase(ValueTraceMergeBase):
     __slots__ = ("loop_node", "type_shapes", "type_shape", "recursion")
@@ -559,6 +679,10 @@ class ValueTraceLoopComplete(ValueTraceLoopBase):
     def getTruthValue():
         return None
 
+    @staticmethod
+    def getComparisonValue():
+        return None
+
 
 class ValueTraceLoopIncomplete(ValueTraceLoopBase):
     __slots__ = ()
@@ -584,3 +708,7 @@ class ValueTraceLoopIncomplete(ValueTraceLoopBase):
     @staticmethod
     def getTruthValue():
         return None
+
+    @staticmethod
+    def getComparisonValue():
+        return None
