@@ -17,14 +17,11 @@
 #
 """ Common test infrastructure functions. To be used by test runners. """
 
-from __future__ import print_function
-
 import ast
 import atexit
 import gc
 import hashlib
 import os
-import re
 import shutil
 import signal
 import sys
@@ -35,7 +32,6 @@ from contextlib import contextmanager
 from optparse import OptionGroup, OptionParser
 
 from nuitka.__past__ import subprocess
-from nuitka.freezer.DependsExe import getDependsExePath
 from nuitka.PythonVersions import (
     getPartiallySupportedPythonVersions,
     getSupportedPythonVersions,
@@ -43,21 +39,19 @@ from nuitka.PythonVersions import (
 from nuitka.Tracing import OurLogger, my_print
 from nuitka.tree.SourceReading import readSourceCodeFromFilename
 from nuitka.utils.AppDirs import getCacheDir
-from nuitka.utils.Execution import (
-    check_output,
-    getNullInput,
-    isExecutableCommand,
-    withEnvironmentVarOverriden,
-)
+from nuitka.utils.Execution import check_output, getNullInput, getNullOutput
 from nuitka.utils.FileOperations import (
     areSamePaths,
     getExternalUsePath,
     getFileContentByLine,
     getFileContents,
     getFileList,
+    isPathBelowOrSameAs,
     makePath,
+    openTextFile,
     removeDirectory,
 )
+from nuitka.utils.Jinja2 import getTemplate
 from nuitka.utils.Utils import getOS
 
 from .SearchModes import (
@@ -146,10 +140,10 @@ print("Anaconda" if os.path.exists(os.path.join(sys.prefix, 'conda-meta')) else
     _python_vendor = version_output.split(b"\n")[3].strip()
 
     if str is not bytes:
-        _python_version_str = _python_version_str.decode("utf-8")
-        _python_arch = _python_arch.decode("utf-8")
-        _python_executable = _python_executable.decode("utf-8")
-        _python_vendor = _python_vendor.decode("utf-8")
+        _python_version_str = _python_version_str.decode("utf8")
+        _python_arch = _python_arch.decode("utf8")
+        _python_executable = _python_executable.decode("utf8")
+        _python_vendor = _python_vendor.decode("utf8")
 
     assert type(_python_version_str) is str, repr(_python_version_str)
     assert type(_python_arch) is str, repr(_python_arch)
@@ -209,11 +203,9 @@ def convertUsing2to3(path, force=False):
     command = [os.environ["PYTHON"], "-m", "py_compile", path]
 
     if not force:
-        with open(path) as source_file:
-            if "xrange" not in source_file.read():
-                with open(os.devnull, "w") as stderr:
-                    if check_result(command, stderr=stderr):
-                        return path, False
+        if "xrange" not in getFileContents(path):
+            if check_result(command, stderr=getNullOutput()):
+                return path, False
 
     filename = os.path.basename(path)
 
@@ -242,22 +234,20 @@ def convertUsing2to3(path, force=False):
 
     command += ("-w", "-n", "--no-diffs", new_path)
 
-    with open(os.devnull, "w") as devnull:
-        try:
-            check_output(command, stderr=devnull)
+    try:
+        check_output(command, stderr=getNullOutput())
 
-        except subprocess.CalledProcessError:
-            if os.name == "nt":
-                raise
+    except subprocess.CalledProcessError:
+        if os.name == "nt":
+            raise
 
-            command[0:3] = ["2to3"]
+        command[0:3] = ["2to3"]
 
-            check_output(command, stderr=devnull)
+        check_output(command, stderr=getNullOutput())
 
-    with open(new_path) as result_file:
-        data = result_file.read()
+    data = getFileContents(new_path)
 
-    with open(new_path, "w") as result_file:
+    with openTextFile(new_path, "w") as result_file:
         result_file.write("__file__ = %r\n" % os.path.abspath(path))
         result_file.write(data)
 
@@ -334,12 +324,26 @@ def decideFilenameVersionSkip(filename):
     if filename.endswith("39.py") and _python_version < (3, 9):
         return False
 
+    # Skip tests that require Python 3.10 at least.
+    if filename.endswith("310.py") and _python_version < (3, 10):
+        return False
+
     return True
 
 
 def decideNeeds2to3(filename):
     return _python_version >= (3,) and not filename.endswith(
-        ("32.py", "33.py", "34.py", "35.py", "36.py", "37.py", "38.py", "39.py")
+        (
+            "32.py",
+            "33.py",
+            "34.py",
+            "35.py",
+            "36.py",
+            "37.py",
+            "38.py",
+            "39.py",
+            "310.py",
+        )
     )
 
 
@@ -449,8 +453,7 @@ def checkCompilesNotWithCPython(dirname, filename, search_mode):
 def checkSucceedsWithCPython(filename):
     command = [_python_executable, filename]
 
-    with open(os.devnull, "w") as devnull:
-        result = subprocess.call(command, stdout=devnull, stderr=subprocess.STDOUT)
+    result = subprocess.call(command, stdout=getNullOutput(), stderr=subprocess.STDOUT)
 
     return result == 0
 
@@ -499,338 +502,12 @@ def displayRuntimeTraces(logger, path):
             os.system("strace -s4096 -e file %s" % path)
 
 
-def getRuntimeTraceOfLoadedFiles(logger, path):
-    """Returns the files loaded when executing a binary."""
-
-    # This will make a crazy amount of work,
-    # pylint: disable=I0021,too-many-branches,too-many-locals,too-many-statements
-
-    if not os.path.exists(path):
-        # TODO: Have a logger package passed.
-        logger.sysexit("Error, cannot find %r (%r)." % (path, os.path.abspath(path)))
-
-    result = []
-
-    if os.name == "posix":
-        if getOS() in ("Darwin", "FreeBSD"):
-            if not isExecutableCommand("dtruss"):
-                test_logger.sysexit(
-                    """\
-Error, needs 'dtruss' on your system to scan used libraries."""
-                )
-
-            if not isExecutableCommand("sudo"):
-                test_logger.sysexit(
-                    """\
-Error, needs 'sudo' on your system to scan used libraries."""
-                )
-
-            args = ("sudo", "dtruss", "-t", "open", os.path.abspath(path))
-        else:
-            if not isExecutableCommand("strace"):
-                test_logger.sysexit(
-                    """\
-Error, needs 'strace' on your system to scan used libraries."""
-                )
-
-            args = (
-                "strace",
-                "-e",
-                "file",
-                "-s4096",  # Some paths are truncated in output otherwise
-                os.path.abspath(path),
-            )
-
-        # Ensure executable is not polluted with third party stuff,
-        # tests may fail otherwise due to unexpected libs being loaded
-        with withEnvironmentVarOverriden("LD_PRELOAD", None):
-            tracing_command = args[0] if args[0] != "sudo" else args[1]
-
-            process = subprocess.Popen(
-                args=args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
-            )
-
-            _stdout_strace, stderr_strace = process.communicate()
-            exit_strace = process.returncode
-
-            if exit_strace != 0:
-                if str is not bytes:
-                    stderr_strace = stderr_strace.decode("utf8")
-
-                logger.warning(stderr_strace)
-                logger.sysexit("Failed to run %r." % tracing_command)
-
-            with open(path + ".strace", "wb") as f:
-                f.write(stderr_strace)
-
-            for line in stderr_strace.split(b"\n"):
-                if process.returncode != 0:
-                    logger.my_print(line)
-
-                if not line:
-                    continue
-
-                # Don't consider files not found. The "site" module checks lots
-                # of things.
-                if b"ENOENT" in line:
-                    continue
-
-                if line.startswith(b"stat(") and b"S_IFDIR" in line:
-                    continue
-
-                # Allow stats on the python binary, and stuff pointing to the
-                # standard library, just not uses of it. It will search there
-                # for stuff.
-                if (
-                    line.startswith(b"lstat(")
-                    or line.startswith(b"stat(")
-                    or line.startswith(b"readlink(")
-                ):
-                    filename = line[line.find(b"(") + 2 : line.find(b", ") - 1]
-
-                    # At least Python3.7 considers the default Python3 path.
-                    if filename == b"/usr/bin/python3":
-                        continue
-
-                    if filename in (
-                        b"/usr/bin/python3." + version
-                        for version in (b"5", b"6", b"7", b"8", b"9")
-                    ):
-                        continue
-
-                    binary_path = _python_executable
-                    if str is not bytes:
-                        binary_path = binary_path.encode("utf-8")
-
-                    found = False
-                    while binary_path:
-                        if filename == binary_path:
-                            found = True
-                            break
-
-                        if binary_path == os.path.dirname(binary_path):
-                            break
-
-                        binary_path = os.path.dirname(binary_path)
-
-                        if filename == os.path.join(
-                            binary_path,
-                            b"python"
-                            + (
-                                "%d%d" % (_python_version[0], _python_version[1])
-                            ).encode("utf8"),
-                        ):
-                            found = True
-                            continue
-
-                    if found:
-                        continue
-
-                result.extend(
-                    os.path.abspath(match)
-                    for match in re.findall(b'"(.*?)(?:\\\\0)?"', line)
-                )
-
-            if sys.version.startswith("3"):
-                result = [s.decode("utf-8") for s in result]
-    elif os.name == "nt":
-        command = (
-            getDependsExePath(),
-            "-c",  # Console mode
-            "-ot%s" % path + ".depends",
-            "-f1",
-            "-pb",
-            "-pa1",  # Turn on all profiling options.
-            "-ps1",  # Simulate ShellExecute with app dirs in PATH.
-            "-pp1",  # Do not long DllMain calls.
-            "-po1",  # Log DllMain call for all other messages.
-            "-ph1",  # Hook the process.
-            "-pl1",  # Log LoadLibrary calls.
-            "-pt1",  # Thread information.
-            "-pe1",  # First chance exceptions.
-            "-pg1",  # Log GetProcAddress calls.
-            "-pf1",  # Use full paths.
-            "-pc1",  # Profile child processes.
-            path,
-        )
-
-        # TODO: Move the handling of this into nuitka.tools.Execution module methods.
-        try:
-            subprocess.call(command, timeout=5 * 60)
-        except Exception as e:  # Catch all the things, pylint: disable=broad-except
-            if e.__class__.__name__ == "TimeoutExpired":
-                test_logger.warning(
-                    "Timeout encountered when running dependency walker."
-                )
-                return []
-            else:
-                raise
-
-        inside = False
-        for line in getFileContentByLine(path + ".depends"):
-            if "| Module Dependency Tree |" in line:
-                inside = True
-                continue
-
-            if not inside:
-                continue
-
-            if "| Module List |" in line:
-                break
-
-            if "]" not in line:
-                continue
-
-            # Skip missing DLLs, apparently not needed anyway.
-            if "?" in line[: line.find("]")]:
-                continue
-
-            dll_filename = line[line.find("]") + 2 :].rstrip()
-            dll_filename = os.path.normcase(dll_filename)
-
-            assert os.path.isfile(dll_filename), repr(dll_filename)
-
-            # The executable itself is of course exempted.
-            if dll_filename == os.path.normcase(os.path.abspath(path)):
-                continue
-
-            result.append(dll_filename)
-
-        os.unlink(path + ".depends")
-
-    result = list(sorted(set(result)))
-
-    return result
-
-
-def checkRuntimeLoadedFilesForOutsideAccesses(loaded_filenames, white_list):
-    # A lot of special white listing is required.
-    # pylint: disable=too-many-branches,too-many-statements
-
-    result = []
-
-    for loaded_filename in loaded_filenames:
-        loaded_filename = os.path.normpath(loaded_filename)
-        loaded_filename = os.path.normcase(loaded_filename)
-        loaded_basename = os.path.basename(loaded_filename)
-
-        ok = False
-        for entry in white_list:
-            if loaded_filename.startswith(entry):
-                ok = True
-
-            while entry:
-                old_entry = entry
-                entry = os.path.dirname(entry)
-
-                if old_entry == entry:
-                    break
-
-                if loaded_filename == entry:
-                    ok = True
-                    break
-        if ok:
-            continue
-
-        if loaded_filename.startswith(("/etc/", "/usr/etc", "/usr/local/etc")):
-            continue
-
-        if loaded_filename.startswith("/proc/") or loaded_filename == "/proc":
-            continue
-
-        if loaded_filename.startswith("/dev/"):
-            continue
-
-        if loaded_filename.startswith("/tmp/"):
-            continue
-
-        if loaded_filename.startswith("/run/"):
-            continue
-
-        if loaded_filename.startswith("/sys/"):
-            continue
-
-        if loaded_filename.startswith("/usr/lib/locale/"):
-            continue
-
-        if loaded_filename.startswith("/usr/share/locale/"):
-            continue
-
-        if loaded_filename.startswith("/usr/share/X11/locale/"):
-            continue
-
-        # Themes may of course be loaded.
-        if loaded_filename.startswith("/usr/share/themes"):
-            continue
-        if "gtk" in loaded_filename and "/engines/" in loaded_filename:
-            continue
-
-        # Terminal info files are OK too.
-        if loaded_filename.startswith("/lib/terminfo/"):
-            continue
-
-        # System C libraries are to be expected.
-        if loaded_basename.startswith(
-            (
-                "ld-linux-x86-64.so",
-                "libc.so.",
-                "libpthread.so.",
-                "libm.so.",
-                "libdl.so.",
-                "libBrokenLocale.so.",
-                "libSegFault.so",
-                "libanl.so.",
-                "libcidn.so.",
-                "libcrypt.so.",
-                "libmemusage.so",
-                "libmvec.so.",
-                "libnsl.so.",
-                "libnss_compat.so.",
-                "libnss_db.so.",
-                "libnss_dns.so.",
-                "libnss_files.so.",
-                "libnss_hesiod.so.",
-                "libnss_nis.so.",
-                "libnss_nisplus.so.",
-                "libpcprofile.so",
-                "libresolv.so.",
-                "librt.so.",
-                "libthread_db-1.0.so",
-                "libthread_db.so.",
-                "libutil.so.",
-            )
-        ):
-            continue
-
-        # Taking these from system is harmless and desirable
-        if loaded_basename.startswith(("libz.so", "libgcc_s.so")):
-            continue
-
-        # TODO: Unclear, loading gconv from filesystem of installed system
-        # may be OK or not. I think it should be.
-        if loaded_basename == "gconv-modules.cache":
-            continue
-        if "/gconv/" in loaded_filename:
-            continue
-        if loaded_basename.startswith("libicu"):
-            continue
-
-        # GTK may access X files.
-        if loaded_basename == ".Xauthority":
-            continue
-
-        result.append(loaded_filename)
-
-    return result
-
-
 def hasModule(module_name):
-    with open(os.devnull, "w") as devnull:
-        result = subprocess.call(
-            (os.environ["PYTHON"], "-c", "import %s" % module_name),
-            stdout=devnull,
-            stderr=subprocess.STDOUT,
-        )
+    result = subprocess.call(
+        (os.environ["PYTHON"], "-c", "import %s" % module_name),
+        stdout=getNullOutput(),
+        stderr=subprocess.STDOUT,
+    )
 
     return result == 0
 
@@ -950,7 +627,7 @@ def checkReferenceCount(checked_function, max_rounds=20, explain=False):
     # Clean start conditions.
     assert sys.exc_info() == (None, None, None), sys.exc_info()
 
-    print(checked_function.__name__ + ": ", end="")
+    my_print(checked_function.__name__ + ": ", end="")
     sys.stdout.flush()
 
     disablePrinting()
@@ -989,29 +666,32 @@ def checkReferenceCount(checked_function, max_rounds=20, explain=False):
     reenablePrinting()
 
     if result:
-        print("PASSED")
+        my_print("PASSED")
     else:
-        print("FAILED", ref_count1, ref_count2, "leaked", ref_count2 - ref_count1)
+        my_print(
+            "FAILED %d %d leaked %d" % (ref_count1, ref_count2, ref_count2 - ref_count1)
+        )
 
         if explain:
             print("REPORT of differences:")
             assert m1
             assert m2
 
+            # Using items will unwanted usages, pylint: disable=consider-using-dict-items
             for key in m1:
                 if key not in m2:
-                    print("*" * 80)
-                    print("extra:", m1[key], key)
+                    my_print("*" * 80)
+                    my_print("extra:", m1[key], key)
                 elif m1[key] != m2[key]:
-                    print("*" * 80)
-                    print(m1[key], "->", m2[key], key)
+                    my_print("*" * 80)
+                    my_print(m1[key], "->", m2[key], key)
                 else:
                     pass
 
             for key in m2:
                 if key not in m1:
-                    print("*" * 80)
-                    print("missing:", m2[key], key)
+                    my_print("*" * 80)
+                    my_print("missing:", m2[key], key)
 
                     # print m1[key]
 
@@ -1143,7 +823,7 @@ def executeReferenceChecked(prefix, names, tests_skipped, tests_stderr, explain=
         # Avoid unraisable output.
         try:
             if number in tests_stderr:
-                sys.stderr = open(os.devnull, "wb")
+                sys.stderr = getNullOutput()
         except OSError:  # Windows
             if not checkReferenceCount(names[name], explain=explain):
                 result = False
@@ -1195,13 +875,6 @@ def withPythonPathChange(python_path):
             old_path = None
             os.environ["PYTHONPATH"] = python_path
 
-    #     print(
-    #         "Effective PYTHONPATH in %s is %r" % (
-    #             sys.modules["__main__"],
-    #             os.environ.get("PYTHONPATH", "")
-    #         )
-    #     )
-
     yield
 
     if python_path:
@@ -1494,9 +1167,6 @@ def scanDirectoryForTestCases(dirname, template_context=None):
         or filename.endswith(".j2")
     ]
 
-    # Jinja2 environment is optional.
-    env = None
-
     for filename in sorted(filenames):
         if not decideFilenameVersionSkip(filename):
             continue
@@ -1505,22 +1175,14 @@ def scanDirectoryForTestCases(dirname, template_context=None):
             # Needs to be a dictionary with template arguments.
             assert template_context is not None
 
-            if env is None:
-                import jinja2
-
-                env = jinja2.Environment(
-                    loader=jinja2.FileSystemLoader("."),
-                    trim_blocks=True,
-                    lstrip_blocks=True,
-                )
-                env.undefined = jinja2.StrictUndefined
-
-            template = env.get_template(filename)
+            template = getTemplate(
+                package_name=None, template_name=filename, template_subdir=dirname
+            )
 
             code = template.render(name=template.name, **template_context)
 
             filename = filename[:-3]
-            with open(filename, "w") as output:
+            with openTextFile(filename, "w") as output:
                 output.write(
                     "'''Automatically generated test, not part of releases or git.\n\n'''\n"
                 )
@@ -1602,16 +1264,15 @@ def checkRequirements(filename):
         if line.startswith("# nuitka-skip-unless-"):
             if line[21:33] == "expression: ":
                 expression = line[33:]
-                with open(os.devnull, "w") as devnull:
-                    result = subprocess.call(
-                        (
-                            os.environ["PYTHON"],
-                            "-c",
-                            "import sys, os; sys.exit(not bool(%s))" % expression,
-                        ),
-                        stdout=devnull,
-                        stderr=subprocess.STDOUT,
-                    )
+                result = subprocess.call(
+                    (
+                        os.environ["PYTHON"],
+                        "-c",
+                        "import sys, os; sys.exit(not bool(%s))" % expression,
+                    ),
+                    stdout=getNullOutput(),
+                    stderr=subprocess.STDOUT,
+                )
                 if result != 0:
                     return (False, "Expression '%s' evaluated to false" % expression)
 
@@ -1710,31 +1371,32 @@ def checkLoadedFileAccesses(loaded_filenames, current_dir):
         if loaded_filename.startswith(current_dir_ext):
             continue
 
-        if loaded_filename.startswith("/etc/"):
-            continue
-
-        if loaded_filename.startswith("/usr/etc/"):
-            continue
-
-        if loaded_filename.startswith("/proc/") or loaded_filename == "/proc":
-            continue
-
-        if loaded_filename.startswith("/dev/"):
-            continue
-
-        if loaded_filename.startswith("/tmp/"):
-            continue
-
-        if loaded_filename.startswith("/run/"):
-            continue
-
-        if loaded_filename.startswith("/usr/lib/locale/"):
-            continue
-
-        if loaded_filename.startswith("/usr/share/locale/"):
-            continue
-
-        if loaded_filename.startswith("/usr/share/X11/locale/"):
+        ignore = True
+        for ignored_dir in (
+            # System configuration is OK
+            "/etc",
+            "/usr/etc",
+            "/usr/local/etc",
+            # Runtime user state and kernel information is OK.
+            "/proc",
+            "/dev",
+            "/run",
+            "/sys",
+            "/tmp",
+            "/var",
+            # Locals may of course be loaded.
+            "/usr/lib/locale",
+            "/usr/share/locale",
+            "/usr/share/X11/locale",
+            # Themes may of course be loaded.
+            "/usr/share/themes",
+            # Terminal info files are OK too.
+            "/lib/terminfo",
+        ):
+            if isPathBelowOrSameAs(ignored_dir, loaded_filename):
+                ignore = False
+                break
+        if not ignore:
             continue
 
         # Themes may of course be loaded.
@@ -1911,7 +1573,38 @@ def checkLoadedFileAccesses(loaded_filenames, current_dir):
         if loaded_filename.startswith("/var/cache/"):
             continue
 
-        _python_version = tuple(int(d) for d in _python_version_str.split("."))
+        # At least Python3.7 considers the default Python3 path and checks it.
+        if loaded_filename == "/usr/bin/python3":
+            continue
+
+        # Accessing the versioned Python3.x binary is also happening.
+        if loaded_filename in (
+            "/usr/bin/python3." + version for version in ("5", "6", "7", "8", "9", "10")
+        ):
+            continue
+
+        binary_path = _python_executable
+
+        found = False
+        while binary_path:
+            if loaded_filename == binary_path:
+                found = True
+                break
+
+            if binary_path == os.path.dirname(binary_path):
+                break
+
+            binary_path = os.path.dirname(binary_path)
+
+            if loaded_filename == os.path.join(
+                binary_path,
+                "python" + ("%d%d" % (_python_version[0], _python_version[1])),
+            ):
+                found = True
+                break
+
+        if found:
+            continue
 
         lib_prefix_dir = "/usr/lib/python%d.%s" % (
             _python_version[0],
@@ -2008,3 +1701,15 @@ def checkLoadedFileAccesses(loaded_filenames, current_dir):
         illegal_accesses.append(orig_loaded_filename)
 
     return illegal_accesses
+
+
+def getMainProgramFilename(filename):
+    for filename_main in os.listdir(filename):
+        if filename_main.endswith(("Main.py", "Main")):
+            return filename_main
+
+    test_logger.sysexit(
+        """\
+Error, no file ends with 'Main.py' or 'Main' in %s, incomplete test case."""
+        % (filename)
+    )
